#include <libce.h>
#include <apic.h>
#include <hd.h>
#include <bm.h>

using namespace hd;

void hd_enable_intr()
{
	outputb(0x00, iDCR);
}

void hd_ack_intr()
{
	inputb(iSTS);
}

namespace bm {

	ioaddr iBMCMD;
	ioaddr iBMSTR;
	ioaddr iBMDTPR;

}

using namespace bm;

bool bm_find(natb& bus, natb& dev, natb& fun)
{
	natb code[] = { 0xff, 0x01, 0x01 };

	return pci_find_class(bus, dev, fun, code);
}

void bm_init(natb bus, natb dev, natb fun)
{
	natl base = pci_read_confl(bus, dev, fun, 0x20);
	base &= ~0x1;
	iBMCMD  = static_cast<ioaddr>(base + 0);
	iBMSTR  = static_cast<ioaddr>(base + 2);
	iBMDTPR = static_cast<ioaddr>(base + 4);
	natw cmd = pci_read_confw(bus, dev, fun, 4);
	pci_write_confw(bus, dev, fun, 4, cmd | 0b101);
}

void bm_prepare(natq prd, bool write)
{
	outputl(prd, iBMDTPR);
	natb work = inputb(iBMCMD);
	if (write) {
		work &= ~0x8;
	} else {
		work |= 0x8;
	}
	outputb(work, iBMCMD);
	work = inputb(iBMSTR);
	work |= 0x6;
	outputb(work, iBMSTR);
}

void bm_start()
{
	natb work = inputb(iBMCMD);
	work |= 1;
	outputb(work, iBMCMD);
}

void bm_ack()
{
	natb work = inputb(iBMCMD);
	work &= 0xFE;
	outputb(work, iBMCMD);
	inputb(iBMSTR);
}

volatile bool done = false;
extern char vv[];
const natl BUFSIZE = 65536;
extern natl prd[];

extern "C" void a_bmide();
extern "C" void c_bmide()
{
	done = true;
	bm_ack();
	hd_ack_intr();
	apic_send_EOI();
}
const natw HD_VECT = 0x60;

int main()
{
	natb nn = BUFSIZE / 512;
	natb lba = 0;
	natb bus = 0, dev = 0, fun = 0;

	clear_screen(0x0f);

	if (!bm_find(bus, dev, fun))
		printf("bm non trovato!\n");
	printf("PCI-ATA at %2x:%2x:%2x\n", bus, dev, fun);
	bm_init(bus, dev, fun);

	apic_set_VECT(14, HD_VECT);
	gate_init(HD_VECT, a_bmide);
	apic_set_MIRQ(14, false);

	for (int i = 0; i < BUFSIZE; i++)
		vv[i] = '-';

	printf("primi 80 caratteri di vv:\n");
	for (int i = 0; i < 80; i++)
		char_write(vv[i]);
	printf("ultimi 80 caratteri di vv:\n");
	for (int i = BUFSIZE - 80; i < BUFSIZE; i++)
		char_write(vv[i]);

	//char *buf = (char *)((natq)&vv & 0xFFFFFFFFFFFF0000);
	//printf("80 byte all'indirizzo %p\n", buf);
	//for (int i = 0; i < 80; i++)
	//	char_write(buf[i]);

	prd[0] = reinterpret_cast<natq>(vv);
	prd[1] = 0x80000000 | ((nn * 512) & 0xFFFF);
	bm_prepare(reinterpret_cast<natq>(prd), false);

	hd_enable_intr();
	hd_start_cmd(lba, nn, READ_DMA);
	bm_start();

	printf("aspetto l'interrupt...\n");
	while (!done)
		;

	printf("primi 80 caratteri di vv:\n");
	for (int i = 0; i < 80; i++)
		char_write(vv[i]);

	printf("ultimi 80 caratteri di vv:\n");
	for (int i = BUFSIZE - 80; i < BUFSIZE; i++)
		char_write(vv[i]);

	//printf("80 byte all'indirizzo %p\n", buf);
	//for (int i = 0; i < 80; i++)
	//	char_write(buf[i]);
	pause();
}
